"""Facebook API"""
import os
import json
import logging
import pandas as pd
import time
import numpy as np
from datetime import datetime
from facebook_business.api import FacebookAdsApi
from facebook_business.adobjects.adaccount import AdAccount
from facebook_business.adobjects.serverside.event import Event
from facebook_business.adobjects.serverside.event_request import EventRequest
from facebook_business.adobjects.serverside.user_data import UserData
from facebook_business.adobjects.serverside.custom_data import CustomData
from facebook_business.exceptions import FacebookRequestError


def transform_campaign_budget(campaigns):
    """
    Transforms get_campaigns response.
    """
    out = []
    for campaign in campaigns:
        campaign_dict = dict(campaign)
        if "lifetime_budget" in campaign_dict:
            campaign_dict["budget"] = campaign_dict["lifetime_budget"]
            campaign_dict["budget_type"] = "lifetime_budget"
            del campaign_dict["lifetime_budget"]
        elif "daily_budget" in campaign_dict:
            campaign_dict["budget"] = campaign_dict["daily_budget"]
            campaign_dict['budget_type'] = "daily_budget"
            del campaign_dict["daily_budget"]
        out.append(campaign_dict)

    data = pd.DataFrame(out).rename(
        columns={
            "id": "campaign_id",
            "name": "campaign_name"
        }
    )
    return data


def build_predicted_revenue_events(df, event_name):
    """
    Creates a list of Facebook Event objects which can be pushed to the Facebook Conversions API.
    Also creates DataFrame for logging which can be used to stream insert to a BigQuery log table.

    :param df: A DataFrame with the events to build Facebook events for
    :type df: pd.DataFrame

    Returns: A tuple with a list of Facebook events and a DataFrame for logs
    rtype: (list of Event, pd.DataFrame)
    """
    events = []
    logs = []
    for index, row in df.iterrows():
        date = int(time.mktime(datetime.strptime(row['date'], '%Y%m%d').timetuple()))
        user_data = UserData(
            country_code=row['shop'],
            fbp=row['facebook_browser_id']
        )

        custom_data = CustomData(
            currency=row['currency'],
            value=row['predicted_revenue']
        )

        event = Event(
            event_name=event_name,
            event_time=date,
            user_data=user_data,
            custom_data=custom_data,
            data_processing_options=[]
        )
        events.append(event)

        logs.append(
            {
                "facebook_browser_id": row['facebook_browser_id'],
                "shop": row['shop'],
                "date_source": row['date'],
                "date_processed": datetime.today().strftime('%Y-%m-%d-%H:%M:%S'),
                "predicted_revenue": row['predicted_revenue'],
                "currency": row['currency']
            }
        )
    df_logs = pd.DataFrame(logs)
    return events, df_logs


def calculate_batches(total_events, batch_size):
    """
    Calculate number of batches to split events into given a batch size and taking into account evenly divisible batches
    :param total_events: The number of total events
    :type total_events: int
    :param batch_size: The max number of events in a batch
    :type batch_size: int

    Returns: The resulting number of batches
    rtype: int

    """
    batches = None
    try:
        batches = (total_events // batch_size)
        batches = batches + 1 if total_events % batch_size != 0 else batches
    except ZeroDivisionError:
        logging.error('Batch Size cannot be 0')
        raise
    except TypeError:
        logging.error('Total Events and Batch Size must be integers')
        raise
    return batches


def split_events_to_batches(events, batch_size=1000):
    """
    Divides a DataFrame of events into batches of the specified batch size (defaults to 1000).

    :param events: A DataFrame of Facebook Events to push to the conversions API
    :type events: pd.DataFrame
    :param batch_size: The max number of events in a batch
    :type batch_size: int

    Returns: A list of DataFrames
    rtype: list of pd.DataFrame

    """
    batched_df = []
    try:
        total_events = len(events.index)
    except TypeError:
        logging.warning("Events is null")

    if total_events > 0:
        logging.info('Batch limit set to %s', batch_size)
        if total_events > batch_size:
            batches = calculate_batches(total_events, batch_size)
            batched_df = np.array_split(events, batches)
            logging.info('Total %s events split into %s batches', total_events, batches)
        else:
            batched_df = [events]
            logging.info('Total %s events only requires 1 batch', total_events)
    else:
        logging.warning('No events split into batches')
    return batched_df


class FacebookExecutor:
    """ Facebook FacebookExecutor.
    Arguments:
        """
    def __init__(self):
        self.client = None
        self.access_token = None
        self.account = None
        self.account_id = None
        self.pixel_id = None
        self.set_api_config()

    def set_api_config(self):
        """
        Loads access_token from FACEBOOK_APPLICATION_CREDENTIALS.
        """
        try:
            with open(os.environ["FACEBOOK_APPLICATION_CREDENTIALS"]) as facebook_cred:
                data = json.load(facebook_cred)
                self.access_token = data["access_token"]
        except KeyError:
            raise KeyError("FACEBOOK_APPLICATION_CREDENTIALS env variable needed")
        self.set_client()

    def set_client(self):
        """
        Sets self.client using the access token.
        """
        self.client = FacebookAdsApi.init(access_token=self.access_token)

    def set_account(self, account_id):
        """ Sets account object
        """
        self.account_id = account_id
        self.account = AdAccount('act_{}'.format(self.account_id))
        logging.info("Initiated AdAccount object for account %s", self.account_id)

    def set_pixel_id(self, pixel_id):
        """ Sets the Pixel ID
        """
        self.pixel_id = pixel_id
        logging.info("Set the pixel_id as %s", self.pixel_id)

    def get_campaign_insights(self, account_id, fields, start_date, end_date):
        """
        Sets insights from the Facebook Insight API.
        Parameters:
            account_id: ID associated to the Facebook Account
            start_date: The start date
            end_date: The end date
            fields: list of field to be fetched
            start_date/end_date: defines the time range to get insights for (YYYY-mm-dd).
        """
        self.set_account(account_id)
        out = []
        params = {
            'effective_status': ['ACTIVE'],
            'level': 'campaign',
            'time_range': {
                'since': start_date,
                'until': end_date
                }
        }
        logging.debug("Downloading insights for account %s", self.account_id)
        logging.debug("fields: %s", fields)
        logging.debug("params: %s", params)
        campaign_insights = self.account.get_insights(
            params=params,
            fields=fields
        )

        for insight in campaign_insights:
            out.append(dict(insight))
        return out

    def get_active_campaigns(self):
        return self.account.get_campaigns(
            fields=['account_id', 'name', 'daily_budget', 'lifetime_budget'],
            params={
                'effective_status': ["ACTIVE"],
                'is_completed': False
            }
        )

    def get_active_campaign_budgets(self, account_id):
        """
        Fetches active campaign metadata from the Facebook API.
        Returns a dataframe with the following fields:
            - account_id
            - campaign_id
            - campaign_name
            - budget_type (daily_budget or lifetime_budget)
            - budget amount in account currency
        """
        self.set_account(account_id)
        campaigns = self.get_active_campaigns()
        out = transform_campaign_budget(campaigns)
        return out

    def update_daily_budget(self, account_id, campaign_id, new_budget):
        """
        Update the budget on the facebook API
        """
        self.set_account(account_id)
        campaigns = self.get_active_campaigns()
        for campaign in campaigns:
            if campaign.get_id() == campaign_id:
                from pygyver.etl.toolkit import configure_logging
                configure_logging()
                logging.info(
                    "Loading new budget for campaign %s",
                    campaign_id
                )
                logging.info(
                    "Current daily_budget for campaign %s: %s",
                    campaign_id,
                    campaign['daily_budget']
                )
                campaign.api_update(
                    params={'daily_budget': round(new_budget*100)}
                )
                logging.info(
                    "New daily_budget for campaign %s: %s",
                    campaign_id,
                    new_budget
                )

        return campaigns

    def push_conversions_api_events(self, events, test_event_code=None):
        """
        Pushes a list of Events to the Facebook Conversions API.

        :param events: A list of Facebook Events to push to the conversions API
        :type events: list of Event
        :param test_event_code: A test_event_code from Facebook Events Manager to mark these as test events
        :type test_event_code: str

        Returns: A dictionary with the parsed response from the Facebook API
        rtype: dict[str, str]
        """
        if len(events) > 1000:
            logging.error("The maximum number of events that Facebook accepts in a single API call is 1,000. "
                          "Please use the split_events_to_batches() function to split the events into batches")
            raise ValueError

        event_request = EventRequest(
            events=events,
            pixel_id=self.pixel_id,
        )

        # Add the test_event_code if one is given
        if test_event_code:
            event_request.test_event_code = test_event_code

        api_response = {}

        try:
            event_response = event_request.execute()
            logging.info('%s events pushed to Facebook Conversions API', event_response.events_received)
            api_response['status'] = 'API Success'
            api_response['fb_trace_id'] = event_response.fbtrace_id
            api_response['messages'] = '\n'.join(event_response.messages)
            api_response['total_events'] = event_response.events_received

        except FacebookRequestError as e:
            logging.error('There was a Facebook Conversions API error:\n\t%s', e)
            api_response['status'] = 'API Error'
            api_response['fb_trace_id'] = e.body()['error']['fbtrace_id']
            error_message = e.body()['error']['message']
            error_message = ': '.join([error_message, e.body()['error']['error_user_msg']])
            api_response['messages'] = error_message
            api_response['total_events'] = None

        return api_response



"""
Ometria API
http://docs.ometria.com/apis/data_api_v2/
- the env vars for the authentication are stored on the kubernetes cluster
under 'ometria-access-credentials'
- functionality:
    set api credentials
    send custom events
    send contact
"""
import logging
import os
import requests
import datetime
import time

class OmetriaExecutor:
    """
    Ometria API handler.
    Args:
        env: switch between environments in Ometria,
            'staging' - for testing
            'prod_marketing' - for marketing emails
            'prod_service' - for service emails
    Attributes:
        api_endpoint: the base API endpoint
        api_key: required for authentication
        api_headers: header to be included in the request
        payload: the formatted payload to be sent
        response: the response from the API call
        queue_safety_limit: limit at which backoff begins
        contact_payload_limit: maximum number of contact records in a single call
        contact_required_fields: required fields for a contact type call
    Returns:
        OmetriaExecutor object
    """

    def __init__(self, env: str):
        """
        Initiate and collect API credentials.
        """
        self.env = env
        self.api_endpoint = "https://api.ometria.com/v2"
        self.api_key = None
        self.api_headers = None
        self.payload = None
        self.response = None
        self.set_api_credentials()
        self.queue_safety_limit = 10000
        self.contact_payload_limit = 100
        self.contact_required_fields = ('id', '@type', 'email')

    def set_api_credentials(self):
        """
        Collect API credentials depending on the environment.
        """
        # api key
        if self.env == "staging":
            api_key_env_var = "OMETRIA_STAGING_API_KEY"
        elif self.env == "prod_marketing":
            api_key_env_var = "OMETRIA_MARKETING_API_KEY"
        elif self.env == "prod_service":
            api_key_env_var = "OMETRIA_SERVICE_API_KEY"
        else:
            raise KeyError(f"Unknown env - {self.env}")

        if api_key_env_var in os.environ:
            self.api_key = os.getenv(api_key_env_var)
            logging.info("API credentials set")
        else:
            raise KeyError(f"Env var {api_key_env_var} does not exist")

        # headers
        self.api_headers = {
            "X-Ometria-Auth": self.api_key,
            "Content-Type": "application/json"
        }

    def send_custom_events(self):
        """
        Send custom_event type of payload to Ometria, save the API response.
        """
        if self.payload:
            # check if payload length is valid - 100 items per send
            payload_len = len(self.payload)
            if payload_len <= 100:
                # request - not adding retry for POST request
                self.response = requests.post(
                    json=self.payload,
                    url=f"{self.api_endpoint}/push",
                    headers=self.api_headers
                )
                logging.info(f"Sent {payload_len} 'custom_events' items")
                self.payload = None

            else:
                raise ValueError(
                    f"Payload too big - {payload_len}, max 100 items"
                )
        else:
            logging.info("No send - empty payload")

    def send_contact(self, payload: list) -> tuple:
        """
        Send a single contact payload to Ometria, backs off when the remaining queue size hits a safety limit

        Args:
            session: the client session
            payload: list of json records to be uploaded

        Returns:
            Tuple of 3 items: a timestamp of when the payload was sent, a list of the uids in the payload, the response json
        """
        self.validate_contact_payload(payload)
        timestamp = datetime.datetime.now()
        uids = [record['id'] for record in payload]
        response = requests.post(f"{self.api_endpoint}/push", json=payload, headers=self.api_headers)
        
        if response.status_code != 202:
            raise ValueError(f"Server responded with {response.status_code} - {response.reason}")
        
        queue_remaining = int(response.headers['X-QueueSize-Remaining'])
        self.backoff(queue_remaining)
        return timestamp, uids, response.json()

    def backoff(self, queue_remaining: int):
        """
        Slow down the rate at which contact calls are made to Ometria to prevent hitting the queue limit

        Args:
            queue_remaining: the number of free slots remaining in the queue
        """
        breach = self.queue_safety_limit - queue_remaining
        if breach > 0:
            delay_time = int((1000 + breach)/1000) * 10
            logging.info(f"Passed queue safety limit of {self.queue_safety_limit}, remaining queue is {queue_remaining}, waiting {delay_time} seconds")
            time.sleep(delay_time)
            logging.info(f"{delay_time} seconds elapsed, resuming...")

    def validate_contact_payload(self, payload: list) -> list:
        """
        Validates contact payloads before uploading, checks that no chunk is > 100 and the individual json records contain the required fields

        Args:
            payloads: a chunked list (max 100 per chunk) of contact records to be uploaded
        """
        if len(payload) > self.contact_payload_limit:
            raise ValueError(f"Payload chunk has size > {self.contact_payload_limit}")
        for record_idx, record in enumerate(payload):
            try:
                [record[field] for field in self.contact_required_fields]
            except KeyError as e:
                logging.error(f"Contact record {record_idx} is missing a required field: {e}")
                raise



""" Functions used in different places ;) """
import json
import numpy as np
import os
import pandas as pd
import re
import string
from datetime import date, datetime, timedelta
from google.cloud import bigquery
from inspect import stack
from pygyver.etl import gs
from pygyver.etl.toolkit import validate_date
from typing import List, Tuple


def set_write_disposition(write_disposition):
    """ Sets bigquery.WriteDisposition based on write_disposition """
    if write_disposition == 'WRITE_APPEND':
        return bigquery.WriteDisposition.WRITE_APPEND
    elif write_disposition == 'WRITE_EMPTY':
        return bigquery.WriteDisposition.WRITE_EMPTY
    elif write_disposition == 'WRITE_TRUNCATE':
        return bigquery.WriteDisposition.WRITE_TRUNCATE
    else:
        raise KeyError("{} is not a valid write_disposition key".format(write_disposition))


def set_priority(priority):
    """ Sets bigquery.QueryPriority based on write_disposition """
    if priority == 'BATCH':
        return bigquery.QueryPriority.BATCH
    elif priority == 'INTERACTIVE':
        return bigquery.QueryPriority.INTERACTIVE
    else:
        raise KeyError("{} is not a valid priority key".format(priority))


def read_table_schema_from_file(path):
    """
    Read table schema from json file.

    Arguments:
        - path: full path to schema file from folder pipelines
    """
    full_path = os.path.join(os.environ.get("PROJECT_ROOT"), path)
    with open(full_path, 'r') as file_path:
        json_schema = json.load(file_path)
        return bigquery.schema._parse_schema_resource({'fields': json_schema})


def bq_token_file_path():
    """
    Returns GOOGLE_APPLICATION_CREDENTIALS if env is set
    """
    return os.environ.get('GOOGLE_APPLICATION_CREDENTIALS', '')


def bq_token_file_path_exists(token_path):
    """
    Returns True if the file exists, False otherwise
    """
    return os.path.exists(token_path)


def bq_token_file_valid():
    """
    Checks whether the token file is valid.
    """
    token_path = bq_token_file_path()
    if token_path == '':
        raise ValueError(
            "Please set GOOGLE_APPLICATION_CREDENTIALS to the path to the access token."
        )
    elif bq_token_file_path_exists(token_path) is False:
        raise ValueError(
            "Token file could not be found. Please reset your GOOGLE_APPLICATION_CREDENTIALS env var. Current:",
            token_path
        )
    else:
        return True


def bq_use_legacy_sql():
    """
    Returns BIGQUERY_LEGACY_SQL if env is set
    """
    return os.environ.get('BIGQUERY_LEGACY_SQL', 'TRUE')


def bq_default_project():
    """
    Returns BIGQUERY_PROJECT if env is set
    """
    return os.environ.get('BIGQUERY_PROJECT', '')


def bq_default_location():
    """
    Returns BIGQUERY_LOCATION if env is set, otherwise defaults to US
    """
    return os.environ.get('BIGQUERY_LOCATION', 'US')


def bq_default_prod_project():
    """
    Returns BIGQUERY_PROD_PROJECT if env is set or 'copper-actor-127213' if not
    """
    return os.environ.get('BIGQUERY_PROD_PROJECT', 'copper-actor-127213')


def bq_default_dataset():
    """
    Returns BIGQUERY_DATASET if env is set
    """
    return os.environ.get('BIGQUERY_DATASET', '')


def bq_billing_project():
    """
    Returns BIGQUERY_PROJECT if env is set
    """
    return bq_default_project()


def bq_start_date():
    """
    Returns BIGQUERY_START_DATE if env is set. Defaults to '2016-01-01'.
    """
    start_date = os.environ.get('BIGQUERY_START_DATE', '2016-01-01')
    validate_date(
        start_date,
        error_msg="Invalid BIGQUERY_START_DATE: {} should be YYYY-MM-DD".format(start_date)
    )
    return start_date


def bq_end_date():
    """
    Returns BIGQUERY_LEGACY_SQL if env is set. Defaults to Yesterday.
    """
    yesterday = (date.today() - timedelta(days=1)).strftime('%Y-%m-%d')
    end_date = os.environ.get('BIGQUERY_END_DATE', yesterday)
    validate_date(
        end_date,
        error_msg="Invalid BIGQUERY_END_DATE: {} should be YYYY-MM-DD".format(end_date)
    )
    return end_date


def remove_first_slash(word=''):
    if word == '':
        return word
    if word[0] == '/':
        return word[1:]
    return word


# AWS
def s3_default_bucket():
    """
    Returns AWS_S3_BUCKET if env is set
    """
    return os.getenv('AWS_S3_BUCKET', '')


def s3_default_root():
    """
    Returns AWS_S3_ROOT if env is set
    """
    return os.getenv('AWS_S3_ROOT', '')


def extract_args(content, to_extract: str, kwargs={}):
    if len(kwargs) > 0:
        for x in content:
            apply_kwargs(x.get(to_extract), kwargs)
    return [x.get(to_extract, '') for x in content if x.get(to_extract, '') != '']


def apply_kwargs(orig, kwargs):
    if orig is not None:
        for key_, value_ in orig.items():
            for kwargs_key, kwargs_value in kwargs.items():
                if '$' + kwargs_key == value_:
                    orig.update({key_: kwargs_value})


def gcs_default_project():
    """
    Returns GOOGLE_CLOUD_PROJECT if env is set
    """
    return os.environ.get('GCS_PROJECT', '')


def gcs_default_bucket():
    """
    Returns GCS_BUCKET if env is set
    """
    return os.environ.get('GCS_BUCKET', '')


def add_dataset_prefix(obj, dataset_prefix: str, kwargs={}):

    if isinstance(obj, list):
        for i in obj:
            add_dataset_prefix(i, dataset_prefix, kwargs)

    if isinstance(obj, dict):
        for k, v in obj.items():
            if isinstance(v, list):
                for i in v:
                    add_dataset_prefix(i, dataset_prefix, kwargs)

            if isinstance(v, dict):
                apply_kwargs(v, kwargs)
                add_dataset_prefix(v, dataset_prefix, kwargs)
            else:
                if k == 'dataset_id':
                    obj[k] = dataset_prefix + obj[k]


def get_dataset_prefix():
    """
    Call this method from a release script listed in pipeline.yaml
    The dataset_prefix can then be added to target dataset_id's to pass dry run tests

    Returns:
        if release module is run from PipelineExecutor:
            return dataset_prefix (string or None)
        else:
            return None
    """
    parent_locals = [
        f[0].f_locals for f in stack()
        if "pygyver/etl/pipeline.py" in f.filename
        and f.function == "run_python_file"
    ]

    if parent_locals:
        dataset_prefix = parent_locals[0].get('_dataset_prefix', None)
    else:
        dataset_prefix = None

    return dataset_prefix


# google sheet to json schema / sql mock
def _convert_column_to_number(
    col_letter: str
) -> int:
    """
    Convert spreadsheet-style column letter to its corresponding number,
    starting with 0.

    Args:
        col_letter: e.g. A, L, AB

    Returns:
        col_number: e.g. 0, 11, 27
    """
    col_number = 0
    for c in col_letter:
        if c in string.ascii_letters:
            col_number = col_number * 26 + (ord(c.upper()) - ord('A')) + 1
    return col_number - 1


def _extract_cell_range_params(
    cell_range: str
) -> Tuple[int, int, List[int]]:
    """
    Extract params to filter the correct cell range from a table.

    Args:
        cell_range - spreadsheet range e.g. B1:C10

    Returns:
        Tuple[
            skiprows - number rows to skip e.g. 0, 3, 10
            nrows - number of rows to insert e.g. 5, 10, 12
            usecols - list of column numbers to insert e.g [0, 1, 2, 3]
        ]
    """
    # extract column and row values
    try:
        cr_1, cr_2 = cell_range.split(":")
        c_1 = re.findall(r"[a-zA-Z]+", cr_1)[0].upper()
        c_2 = re.findall(r"[a-zA-Z]+", cr_2)[0].upper()
        d_1 = int(re.findall(r"\d+", cr_1)[0])
        d_2 = int(re.findall(r"\d+", cr_2)[0])
        cs = sorted([
            _convert_column_to_number(c_1),
            _convert_column_to_number(c_2)
        ])
    except IndexError:
        raise Exception(
            "Incorrect input, required format is e.g. 'B3:L10'"
        )
    # params for gspread
    skiprows = min([d_1, d_2]) - 1
    nrows = abs(d_1 - d_2)
    usecols = list(
        range(
            cs[0],
            cs[1] + 1
        )
    )
    return skiprows, nrows, usecols


def _extract_table_columns(
    key: str,
    sheet_name: str = "acceptance",
    sheet_index: int = 1
) -> dict:
    """
    Extract the table names and column indices from the full gs df.

    Args:
        key: 44-digit spreadsheet id from the url e.g. '1DR1...k-xDU'
        sheet_name: name of the sheet within the spreadsheet e.g. 'Sheet1'
        sheet_index: zero-based index where the sheet is within the spreadsheet

    Returns:
        table_details: table_name-[col_idx] key-value pairs
    """
    # load gs
    df = gs.load_gs_to_dataframe(
        key=key,
        sheet_name=sheet_name,
        sheet_index=sheet_index,
        evaluate_formulas=True
    )
    df = df.dropna(
        axis=1,
        how="all"
    )
    # extract details
    cols = list(df.columns)
    table_names = [
        x for x in cols if
        ("." in x) & ("Unnamed" not in x)
    ]
    table_idx_1 = [
        cols.index(x) for x in table_names
    ]
    table_idx_2 = (
        table_idx_1 + [len(cols)]
    )[1:]

    table_details = {
        v: list(range(table_idx_1[k], table_idx_2[k]))
        for k, v in enumerate(table_names)
    }
    return table_details


def _extract_table_schema(
    key: str,
    cols: List,
    sheet_name: str = "acceptance",
    sheet_index: int = 1
) -> dict:
    """
    Extract the table schema from the full gs df using the column indices.

    Args:
        key: 44-digit spreadsheet id from the url e.g. '1DR1...k-xDU'
        cols: column indices to insert e.g. [3, 4, 5, 6]
        sheet_name: name of the sheet within the spreadsheet e.g. 'Sheet1'
        sheet_index: zero-based index where the sheet is within the spreadsheet

    Returns:
        schema_dict: column name-sql column type as key-value pairs
            e.g. {'uid': 'STRING', 'clicks': 'INTEGER'}
    """
    schema_df = gs.load_gs_to_dataframe(
        key=key,
        usecols=cols,
        skiprows=1,
        nrows=1,
        sheet_name=sheet_name,
        sheet_index=sheet_index,
        evaluate_formulas=True
    )
    schema_df.columns = [
        x.split(".")[0] for x in schema_df.columns
    ]
    schema_dict = dict(
        zip(
            list(schema_df.columns),
            list(schema_df.iloc[0, :])
        )
    )
    return schema_dict


def _extract_table_data(
    key: str,
    cols: List,
    schema: dict,
    sheet_name: str = "acceptance",
    sheet_index: int = 1
) -> pd.DataFrame:
    """
    Extract the table data from the full gs df using the column indices.

    Args:
        key: 44-digit spreadsheet id from the url e.g. '1DR1...k-xDU'
        cols: column indices to insert e.g. [3, 4, 5, 6]
        schema: column name-python type as key-value pairs
            e.g. {'uid': str, 'clicks': 'Int64'}
        sheet_name: name of the sheet within the spreadsheet e.g. 'Sheet1'
        sheet_index: zero-based index where the sheet is within the spreadsheet

    Returns:
        df: pandas dataframe of the selected cell range
    """
    df = gs.load_gs_to_dataframe(
        key=key,
        usecols=cols,
        dtype=schema,
        names=list(schema.keys()),
        sheet_name=sheet_name,
        sheet_index=sheet_index,
        skiprows=3,
        evaluate_formulas=True
    )
    df = df.dropna(how="all")
    df = df.reset_index(drop=True)

    return df


def _schema_sql_to_bq_compatibility(
    schema_dict: dict
) -> dict:
    """
    Convert sql schema to be compatible with the bq ui.

    Args:
        schema_dict: column name-sql column type as key-value pairs
            e.g. {'uid': 'STRING', 'clicks': 'INTEGER'}

    Returns:
        schema_dict: column name-sql column type as key-value pairs
            e.g. {'uid': 'STRING', 'clicks': 'INT64'}
    """
    for k, v in schema_dict.items():
        if v == "INTEGER":
            schema_dict[k] = "INT64"
        elif v == "FLOAT":
            schema_dict[k] = "FLOAT64"

    return schema_dict


def _extract_schema_json(
    schema_path: str
) -> dict:
    """
    Extract the sql schema from the json as key-value dict pairs.

    Args:
        schema_path: path to the schema file e.g. 'path/file.json'

    Returns:
        schema_dict: column name-sql column type as key-value pairs
            e.g. {'uid': 'STRING', 'clicks': 'INTEGER'}
    """
    with open(schema_path) as json_file:
        json_data = json.load(json_file)

    schema_dict = {x["name"]: x["type"] for x in json_data}

    schema_dict = _schema_sql_to_bq_compatibility(schema_dict)

    return schema_dict


def _schema_sql_to_py(
    schema_sql: dict
) -> dict:
    """
    Convert sql schema to python/pandas format.

    Args:
        schema_sql: column name-sql type as key-value pairs
            e.g. {'uid': 'STRING', 'clicks': 'INTEGER'}

    Returns:
        schema_py: column name-python type as key-value pairs
            e.g. {'uid': str, 'clicks': 'Int64'}

    Notes:
        - defaults to string if type not in the converter dict
    """
    schema_sql_to_py = {
        "BOOLEAN": str,  # cast error if using pandas boolean
        "DATE": datetime,
        "DATETIME": datetime,
        "FLOAT": np.float64,
        "FLOAT64": np.float64,
        "INT64": "Int64",
        "INTEGER": "Int64",  # pandas integer type that's nullable
        "NUMERIC": np.float64,
        "STRING": str,
        "TIME": str,
        "TIMESTAMP": str
    }
    schema_py = {
        k: schema_sql_to_py[v]
        if v in schema_sql_to_py
        else str
        for k, v in schema_sql.items()
    }
    return schema_py


def _field_to_cte(
    idx: int,
    field: tuple,
    field_type: str
) -> str:
    """
    Convert a single field to its string equivalent for a cte.

    Args:
        idx: row number, used to define if it's a first line
        field: DataFrame cell as tuple(column_name, value)
            e.g. tuple('store_code', 'gb')
        field_type: sql data type e.g. 'STRING', 'INTEGER'

    Returns:
        field_cte: field in sql cte format as str
            - first row e.g. "CAST('gb' AS STRING) AS country_code"
            - latter rows e.g. "CAST('gb' AS STRING)"
    """
    # handle nulls
    f_format = "NULL" if pd.isna(field[1]) else f"'{field[1]}'"

    # first line - include column name
    field_cte = f"CAST({f_format} AS {field_type})"
    if idx == 0:
        field_cte = f"{field_cte} AS {field[0]}"

    return "\n\t\t" + field_cte


def _row_to_cte(
    idx: int,
    row: pd.core.series.Series,
    schema_sql: dict
) -> str:
    """
    Convert a DataFrame row to its string equivalent for a cte.

    Args:
        idx: row number, used to define if it's a first line
        row: DataFrame row object
        schema_sql: column name-sql type as key-value pairs
            e.g. {'uid': 'STRING', 'clicks': 'INTEGER'}

    Returns:
        line_cte: row in sql cte format as str
    """
    line = ",".join([
        _field_to_cte(idx, r, schema_sql[r[0]]) for r in row.iteritems()
    ])
    if idx == 0:
        line_cte = f"\tSELECT{line}"
    else:
        line_cte = f"\tUNION ALL\n\tSELECT{line}"

    return line_cte


def _df_to_cte(
    df: pd.DataFrame,
    schema_sql: dict,
    cte_name: str
) -> str:
    """
    Convert a DataFrame to a cte format.

    Args:
        df: DataFrame to convert
        schema_sql: column name-sql type as key-value pairs
            e.g. {'uid': 'STRING', 'clicks': 'INTEGER'}
        cte_name: full name (dataset_id.table_id) of the cte table
            e.g. data.country_product

    Returns:
        cte: the complete cte as str
    """
    lines = "\n".join([
        _row_to_cte(idx, row, schema_sql) for idx, row in df.iterrows()
    ])
    cte = f"`{cte_name}` AS (\n{lines}\n)"

    return cte



""" SQL Alchemy Wrapper """
import os
import logging
import pandas as pd
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool
from sqlalchemy.engine.url import URL
from pygyver.etl.toolkit import read_sql_file


class DBExecutorError(Exception):
    pass


class DBExecutor:
    """ Database handler
    Note:
        Sqlalchemy's built-in drivers should be supported
        Additional drivers can be added using requirements.txt
        Postgres(pg8000) and MySQL(pymysql) are included in tests

    Args:
        name (str): Defines connection using env vars with this prefix.
        url (str): Defines connection using SQLAlchemy database URL format.

    Attributes:
        name (str): Defines connection using env vars with this prefix.
        url (str): Defines connection using SQLAlchemy database URL format.
        safe_url (str): url with password removed for logging
        engine: SQLAlchemy Engine object

    Returns:
        DBExecutor object.

    Examples:
        >>> db = DBExecutor("ERP")
        >>> db = DBExecutor(url="mysql+pymysql://username:password@host:port/database")

    References:
        https://docs.sqlalchemy.org/en/13/core/engines.html#database-urls
    """
    def __init__(self, name=None, url=None):
        if url:
            self.url = url
        else:
            if name:
                self.url = URL(
                    drivername=os.getenv(f"{name}_DRIVERNAME"),
                    username=os.getenv(f"{name}_USERNAME"),
                    password=os.getenv(f"{name}_PASSWORD"),
                    host=os.getenv(f"{name}_HOST"),
                    port=os.getenv(f"{name}_PORT"),
                    database=os.getenv(f"{name}_DATABASE")
                )
            else:
                raise DBExecutorError("Either database name or url must be provided")
        # remove password from url string for logging
        self.safe_url = repr(self.url)

        try:
            self.engine = create_engine(self.url, poolclass=NullPool)
        except Exception as error:
            logging.error(f"Failed to connect to {self.safe_url}: {error}")
            raise error


    def clean_up(self):
        self.engine.dispose()


    def execute_query(self, sql=None, file=None, **kwargs):
        """ Executes a SQL query and loads it as a DataFrame.

        Args:
            sql (str): SQL query (or table name).
            file (str): path to the SQL file relative to PROJECT_ROOT env var.
            **kwargs: can be applied to pass parameters into a SQL file

        Returns:
            Query result as a DataFrame.
        """
        if sql is None:
            if file:
                sql = read_sql_file(file, **kwargs)
            else:
                raise DBExecutorError("Either SQL or file containing the SQL must be provided")

        with self.engine.connect() as con:
            try:
                logging.info(f"Running SQL query on {self.safe_url}")
                df = pd.read_sql(sql, con)
            except Exception as error:
                logging.error(error)
                raise error
            finally:
                con.close()

        return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on: Mon June 17 16:27:12 2019
Updated on: Wed Feb 17 21:50:00 2021

@authors: kzaoui, dankruse1, tommy-watts
"""

import os
import time
import re
import json
from time import sleep
import logging
import requests


class ZendeskDownloader:
    """
    Class to download data from a Zendesk API endpoint specified by params.
    """

    def __init__(self, start_time=None, uri=None, sideload=None, per_page=None, limit=None, fields=None, base_url="https://madecom.zendesk.com/"):
        """
        Construct all the necessary attributes for the ZendeskDownloader instance

        Args:
          uri (str): API endpoint.
          start_time (str, optional): Start time of the request
          limit (int, optional): Per request limit of results returned, optional paramater for some endpoints.
          sideload (str, optional): Additional parameters seperated by &.
          per_page (int, optional): Size of page to fetch from the API.
          base_url (str, optional): Base url of the api call.

        Examples:
          downloader = ZendeskDowloader(
            uri="api/v2/incremental/tickets",
            start_time="2018-12-31 22:45:32",
            sideload="include=comment&include=metric_sets",
            per_page=200,
            pagination=True
          )
        """
        self.user = "analytics@made.com/token"
        self.base_url = base_url
        self.start_time = start_time
        self.sideload = sideload
        self.uri = uri
        self.per_page = per_page
        self.limit = limit
        self.fields = fields
        self.endpoint = None
        self.next_page = None
        self.retry = 0
        self.max_retry = 3
        self.params_list = []
        self.endpoint_params = ''
        self.finished = False
        self.headers = None
        self.set_start_time()
        self.set_auth_token()
        self.set_uri()
        self.set_start_time_param()
        self.set_fields_param()
        self.set_per_page_param()
        self.set_limit_param()
        self.set_sideload_param()
        self.set_endpoint_params()

    def set_start_time(self):
        """Convert `self.start_time` to epoch timestamp.

        Raises:
          ValueError: If `self.start_time` does not follow the format "%Y-%m-%d %H:%M:%S"
        """
        pattern = "%Y-%m-%d %H:%M:%S"
        if self.start_time is not None:
            try:
                logging.info("Start date: %s", self.start_time)
                self.start_time = int(time.mktime(
                    time.strptime(self.start_time, pattern)))
            except ValueError:
                raise ValueError("""Incorrect date format, must be '%Y-%m-%d %H:%M:%S',
                eg. '2018-12-31 22:45:32'""")

    def set_auth_token(self):
        """Load access token to `self.headers`

        Raises:
          KeyError: if env var ZENDESK_ACCESS_KEY or ZOPIM_ACCESS_KEY do not exist
        """
        if self.base_url == 'https://madecom.zendesk.com/':
            self.headers = {"Authorization": f"Bearer {os.environ['ZENDESK_ACCESS_KEY']}"}
        elif self.base_url == 'https://www.zopim.com/':
            self.headers = {"Authorization": f"Bearer {os.environ['ZOPIM_ACCESS_KEY']}"}
        else:
            raise KeyError(
                "Invalid base url, should be one of ['https://madecom.zendesk.com/ ', 'https://www.zopim.com/'")

    def set_uri(self):
        """Format the `self.uri` value for use in the endpoint 

        Raises:
          KeyError: if `self.uri` is None
        """
        if self.uri is not None:
            self.uri = "{}.json".format(self.uri)
        else:
            raise KeyError("URI param required")

    def set_fields_param(self):
        """Format and append fields param to params_list if `self.fields` is not None
        """
        if self.fields:
            param = "fields={}".format(self.fields)
            self.params_list.append(param)

    def set_start_time_param(self):
        """Format and append start_time param to params_list if `self.start_time` is not None
        """
        if self.start_time:
            param = "start_time={}".format(self.start_time)
            self.params_list.append(param)

    def set_per_page_param(self):
        """Format and append per_page param to params_list if `self.per_page` is not None
        """
        if self.per_page:
            param = "per_page={}".format(self.per_page)
            self.params_list.append(param)

    def set_limit_param(self):
        """Append limit param to params_list if `self.limit` is not None
        """
        if self.limit:
            param = "limit={}".format(self.limit)
            self.params_list.append(param)

    def set_sideload_param(self):
        """Format and append sideload param to params_list if `self.sideload` is not None
        """
        if self.sideload:
            param = self.sideload
            self.params_list.append(param)

    def set_endpoint_params(self):
        """Set `self.endpoint_params`

        Iterate through `self.params_list` generating a string of params integrating '?' and '&' where required
        """
        for i, k in enumerate(sorted(self.params_list)):
            if i == 0:
                self.endpoint_params = self.endpoint_params + "?{}".format(k)
            else:
                self.endpoint_params = self.endpoint_params + "&{}".format(k)

    def get_next_endpoint(self):
        """Return the next endpoint if it exists

        Check to see if end of stream is True this is required because some endpoints return the next endpoint even though all records have been returned.
        If end of stream is False then return either the next_page or after_url depending on which key exists in the response
        """
        if self.base_url == 'https://www.zopim.com/' and self.data.get("count") < self.limit:
            return None
        if self.data.get("end_of_stream"):
            return None
        return self.data.get("next_page", self.data.get("after_url"))

    def set_endpoint(self):
        """Set `self.endpoint`. 

        The first call sets the endpoint based on values from the provided uri and params.
        Subsequent calls set the endpoint based on the returned next page value from the previous call (next_page/after_url).
        """
        self.retry = 1
        if not self.endpoint:
            self.endpoint = self.base_url + self.uri + self.endpoint_params
            logging.info("Endpoint generated - {}".format(self.endpoint))
        else:
            next_endpoint = self.get_next_endpoint()
            if next_endpoint:
                self.endpoint = next_endpoint
                logging.info("Endpoint generated - {}".format(self.endpoint))
            else:
                self.finished = True

    def api_call(self):
        """ Make the API request and handle retry based response codes.

        After each request update `self.endpoint` with the next_page endpoint provided in the response.
        If the rate limit is reached (429) process sleeps for period defined in the response.

        Raises:
          ValueError: If there is an error response and max retires has been exhausted  
        """
        response = requests.get(url=self.endpoint, headers=self.headers)
        if response.status_code == 200:
            self.data = json.loads(response.content)
        elif response.status_code == 429:  # timed-out
            sleep_time = int(response.headers['retry-after'])
            logging.warning('Too many requests (429). Waiting %s secs to retry…', sleep_time)
            sleep(sleep_time)
            self.api_call()
        else:
            while self.retry < self.max_retry:
                logging.info("Error {}. Attempt {}/{} failed, retrying after 1 second...".format(
                    response.status_code, self.retry, self.max_retry))
                sleep(1)
                self.retry = self.retry + 1
                self.api_call()
            logging.info("Reached max_retry, raising error")
            raise ValueError("HTTP Error {}".format(response.status_code))

    def download(self):
        """ Download API responses based on provided URI and params
        Returns
          obj: Object of API response data
        """
        self.set_endpoint()
        if not self.finished:
            self.api_call()
            return self.data
        logging.info("No further call to be made. Closing loop.")
        return []



import io
import versioneer
import setuptools


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


long_description = read("README.md")

setuptools.setup(
    name="pygyver",
    version=versioneer.get_version(),
    author="Simone Fiorentini",
    author_email="analytics@made.com",
    description="Data engineering & Data science Pipeline Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/madedotcom/pygyver",
    test_suite="pygyver.tests",
    packages=setuptools.find_packages(),
    install_requires=[
        'boto3>=1.9.218',
        'codecov>=2.1.0',
        'facebook-business>=6.0.0',
        'google-cloud-bigquery>=1.24.0',
        'google-cloud-storage>=1.28.0',
        'gspread>=3.1.0',
        'gspread-dataframe>=3.0.4',
        'moto>=1.3.14',
        'nltk>=3.4.0',
        'numpy>=1.13.3',
        'pandas>=1.0.0',
        'pandas-gbq>=0.9.0',
        'pg8000>=1.14.0',
        'pyarrow>=1.0.0',
        'pymysql>=0.9.3',
        'pytest>=5.3.1',
        'PyYAML>=5.1',
        'sqlalchemy>=1.3.0',
        'tox>=3.15.0',
        'versioneer>=0.18',
        'wheel>=0.33.1'
    ],
    cmdclass=versioneer.get_cmdclass(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



""" Module to ETL data to generate pipelines """
from __future__ import print_function

import os
import copy
import logging
import concurrent.futures
import numpy as np
import string
from random import randint
from pathlib import PurePath
from datetime import date
from importlib.util import spec_from_file_location, module_from_spec
from pygyver.etl.dw import read_sql
from pygyver.etl.lib import apply_kwargs
from pygyver.etl.lib import extract_args
from pygyver.etl.dw import BigQueryExecutor
from pygyver.etl.toolkit import read_yaml_file
from pygyver.etl.lib import bq_default_project
from pygyver.etl.lib import bq_default_prod_project
from pygyver.etl.lib import add_dataset_prefix


def execute_parallel(func, args, message='running task', log=''):
    """
    execute the functions in parallel for each list of parameters passed in args

    Arguments:
    func (function): function as an object
    args (list): args associated to the function
    message (string): message to be displayed
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        future_to_func = {executor.submit(func, **arg): arg for arg in args}
        for future in concurrent.futures.as_completed(future_to_func):
            arg = future_to_func[future]
            try:
                ret = future.result()
            except AssertionError:
                logging.info(f"{message} {arg.get(log,'')}: failed")
                raise AssertionError
            except Exception as exc:
                logging.info('%r generated an exception: %s' % (arg, exc))
                raise Exception
            else:
                logging.info(f"{message} {arg.get(log,'')}")


def extract_unit_test_value(unit_test_list):
    utl = copy.deepcopy(unit_test_list)
    for d in utl:
        sql_file = d.pop('file')
        d["sql"] = extract_unit_test_sql_value(d, sql_file)
        d["cte"] = read_sql(file=d['mock_file'], **d)
        d["file"] = sql_file
    return utl

def extract_unit_test_sql_value(test, sql_file):
    sql= read_sql(file=sql_file)
    sql_parser = string.Formatter()
    elements = sql_parser.parse(sql)
    format_dict={}
    for a, b, c, d in elements:
        if 'mock_{}'.format(b) in test :
            format_dict[b]=test['mock_{}'.format(b)]
        elif b in test:
            format_dict[b]=test[b]
    sql = sql.format(**format_dict)
    return sql

def extract_unit_tests(batch_list=None, kwargs={}):
    """ return the list of unit test: unit test -> file, mock_file, output_table_name(opt) """

    # initiate args and argsmock
    args, args_mock = [], []

    # extracts files paths for unit tests
    for batch in batch_list:
        apply_kwargs(batch, kwargs)
        for table in batch.get('tables', ''):
            if (table.get('create_table', '') != '' or table.get('create_partition_table', '') != '') and table.get('mock_data', '') != '':
                if table.get('create_table', '') != '':
                    args.append(table.get('create_table', ''))
                if table.get('create_partition_table', '') != '':
                    args.append(table.get('create_partition_table', ''))
                args_mock.append(table.get('mock_data', ''))

    for a, b in zip(args, args_mock):
        a.update(b)
    return args


class PipelineExecutor:
    def __init__(self, yaml_file, dry_run=False, *args, **kwargs):
        self.kwargs = kwargs
        self.yaml = read_yaml_file(yaml_file)
        self.dataset_prefix = None
        if dry_run:
            self.dataset_prefix = f'{randint(1, 99999999):08}_'
            add_dataset_prefix(obj=self.yaml, dataset_prefix=self.dataset_prefix, kwargs=self.kwargs)
        self.bq = BigQueryExecutor()
        self.prod_project_id = bq_default_prod_project()


    def remove_dataset(self, dataset_id):
        if self.bq.dataset_exists(dataset_id):
            self.bq.delete_dataset(dataset_id, delete_contents=True)

    def dry_run_clean(self, table_list=''):
        if self.dataset_prefix is not None:
            if bq_default_project() != self.prod_project_id:
                args_dataset = []

                if table_list == '':
                    table_list = self.yaml.get('table_list', '')

                for table in table_list:
                    if table.count('.') == 1:
                        dataset_id = table.split(".")[0]
                    else:
                        dataset_id = table.split(".")[1]
                    dict_ = {
                        "dataset_id": dataset_id
                    }
                    apply_kwargs(dict_, self.kwargs)
                    args_dataset.append(
                        dict_
                    )

                for dataset in args_dataset:
                    value = dataset.get('dataset_id', '')
                    dataset['dataset_id'] = self.dataset_prefix + value

                args_dataset = [dict(t) for t in {tuple(d.items()) for d in args_dataset}]

                if args_dataset != []:
                    execute_parallel(
                        self.remove_dataset,
                        args_dataset,
                        message='delete dataset: ',
                        log='dataset_id'
                    )

    def create_tables(self, batch):
        args = []
        batch_content = batch.get('tables', '')
        args = extract_args(content=batch_content, to_extract='create_table', kwargs=self.kwargs)
        for a in args:
            apply_kwargs(a, self.kwargs)
            a.update({"dataset_prefix": self.dataset_prefix})
        if args != []:
            execute_parallel(
                self.bq.create_table,
                args,
                message='Creating table:',
                log='table_id'
            )

    def create_gs_tables(self, batch):
        args =[]
        batch_content = batch.get('sheets', '')
        args = extract_args(content=batch_content, to_extract='create_gs_table', kwargs=self.kwargs)
        if args == []:
            raise Exception("create_gs_table in YAML file is not well defined")
        execute_parallel(
                self.bq.create_gs_table,
                args,
                message='Creating live Google Sheet connection table in BigQuery:',
                log='table_id'
            )


    def create_partition_tables(self, batch):
        args = []
        batch_content = batch.get('tables', '')
        args = extract_args(
            content=batch_content,
            to_extract='create_partition_table',
            kwargs=self.kwargs
        )
        for a in args:
            apply_kwargs(a, self.kwargs)
            a.update({"dataset_prefix": self.dataset_prefix})
        if args != []:
            execute_parallel(
                self.bq.create_partition_table,
                args,
                message='Creating partition table:',
                log='table_id'
            )

    def load_google_sheets(self, batch):
        args = []
        batch_content = batch.get('sheets', '')
        args = extract_args(batch_content, 'load_google_sheet')
        for a in args:
            apply_kwargs(a, self.kwargs)
        if args == []:
            raise Exception("load_google_sheet in yaml is not well defined")
        execute_parallel(
            self.bq.load_google_sheet,
            args,
            message='Loading table:',
            log='table_id'
        )

    def run_checks(self, batch):
        args, args_pk = [], []
        batch_content = batch.get('tables', '')
        args = extract_args(batch_content, 'create_table')
        args_pk = [x.get('pk', []) for x in batch_content]
        for a, b in zip(args, args_pk):
            a.update({
                "dataset_prefix": self.dataset_prefix,
                "primary_key": b
                }
            )
        execute_parallel(
            self.bq.assert_unique,
            args,
            message='Run pk_check on:',
            log='table_id'
        )

    def run_batch(self, batch):
        """ Executes a batch. """

        if 'tables' in batch:
            if extract_args(batch['tables'], 'create_table'):
                self.create_tables(batch)
                self.run_checks(batch)
            if extract_args(batch['tables'], 'create_partition_table'):
                self.create_partition_tables(batch)

        if 'sheets' in batch:
            if extract_args(batch['sheets'], 'load_google_sheet'):
                self.load_google_sheets(batch)
            if extract_args(batch['sheets'], 'create_gs_table'):
                self.create_gs_tables(batch)


    def run_batches(self):
        batch_list = self.yaml.get('batches', '')
        for batch in batch_list:
            apply_kwargs(batch, self.kwargs)
            self.run_batch(batch)

    def run_python_file(self, python_file):
        # _dataset_prefix string is unused in run_python_file()
        # but it makes PipelineExecutor's dataset_prefix available to the release script, using:
        # from pygyver.etl.lib import get_dataset_prefix
        _dataset_prefix = self.dataset_prefix

        logging.info(f"Running {python_file}")
        module_name = PurePath(python_file).stem
        module_full_path = PurePath(os.getenv("PROJECT_ROOT")) / python_file
        spec = spec_from_file_location(module_name, module_full_path)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)

    def run_releases(self, release_date=date.today().strftime("%Y-%m-%d")):
        release_list = self.yaml.get('releases', [])
        for release in release_list:
            if str(release.get('date', '')) == release_date:
                logging.info(f"Release {release_date}: {release.get('description', '')}")
                for python_file in release.get('python_files', []):
                    self.run_python_file(python_file)

    def run(self):
        self.run_releases()
        self.run_batches()

    def run_unit_tests(self, batch_list=None):
        batch_list = batch_list or self.yaml.get('batches', '')
        list_unit_test = extract_unit_tests(batch_list, self.kwargs)
        args = extract_unit_test_value(list_unit_test)
        if args != []:
            execute_parallel(
                self.bq.assert_acceptance,
                args,
                message='Asserting sql',
                log='file'
            )

    def copy_prod_structure(self, table_list=''):
        args, args_dataset, datasets = [], [], []

        if table_list == '':
            table_list = self.yaml.get('table_list', '')

        for table in table_list:
            if table.count('.') == 1:
                _dict = {
                    "source_project_id" : self.prod_project_id,
                    "source_dataset_id" : table.split(".")[0],
                    "source_table_id": table.split(".")[1],
                    "dest_dataset_id" : self.dataset_prefix + table.split(".")[0],
                    "dest_table_id": table.split(".")[1]
                }
            else:
                _dict = {
                    "source_project_id" : table.split(".")[0],
                    "source_dataset_id" : table.split(".")[1],
                    "source_table_id": table.split(".")[2],
                    "dest_dataset_id" : self.dataset_prefix + table.split(".")[1],
                    "dest_table_id": table.split(".")[2]
                }
            apply_kwargs(_dict, self.kwargs)
            args.append(_dict)

        # extract datasets from table_list
        for table in table_list:
            if table.count('.') == 1:
                datasets.append(self.dataset_prefix + table.split(".")[0])
            else:
                datasets.append(self.dataset_prefix + table.split(".")[1])

        for dataset in np.unique(datasets):
            _dict = {"dataset_id" : dataset}
            apply_kwargs(_dict, self.kwargs)
            args_dataset.append(
                _dict
            )

        if args_dataset != []:
            execute_parallel(
                self.bq.create_dataset,
                args_dataset,
                message='create dataset for: ',
                log='dataset_id'
            )

        if args != []:
            execute_parallel(
                self.bq.copy_table_structure,
                args,
                message='copy table structure for: ',
                log='source_table_id'
            )

    def run_test(self):
        self.run_unit_tests()



""" Google Spreadsheet utility """
import gspread
import json
import pandas as pd
from google.auth.transport.requests import AuthorizedSession
from google.oauth2 import service_account
from gspread_dataframe import get_as_dataframe
from pygyver.etl import lib
from pygyver.etl.toolkit import (
    create_folder,
    split_full_path,
    write_text_file
)
from time import sleep


def gspread_client():
    """ Sets BigQuery client.
    """
    lib.bq_token_file_valid()
    credentials = service_account.Credentials.from_service_account_file(
        lib.bq_token_file_path(),
        scopes=[
            'https://spreadsheets.google.com/feeds',
            'https://www.googleapis.com/auth/drive'
        ]
    )
    gc = gspread.Client(auth=credentials)
    gc.session = AuthorizedSession(credentials)
    return gc


def load_gs_to_dataframe(key, sheet_name='', sheet_index=0, **kwargs):
    '''
    Loads Google Spreadsheet to a pandas dataframe

    Args:
        key (str): Google Sheets key a.k.a. Spreadsheet ID
        sheet_name (str): Name of the sheet within the Spreadsheet
        sheet_index (int): Zero-based index where the sheet is within the Spreadsheet
        **kwargs: Any that are supported by this Pandas version’s text parsing readers, such as pandas.read_csv
            e.g. delimiter=None, header='infer', names=None, index_col=None, etc.
            For more info, see docs for gspread-dataframe and pandas.read_csv

    Returns:
        Spreadsheet as a DataFrame.
    '''
    client = gspread_client()

    if sheet_name != '':
        sheet = client.open_by_key(key).worksheet(sheet_name)
    else:
        sheet = client.open_by_key(key).get_worksheet(sheet_index)
    data = get_as_dataframe(sheet, **kwargs)
    return data


def load_gs_range_to_dataframe(
    key: str,
    cell_range: str,
    sheet_name: str = '',
    sheet_index: int = 0,
    **kwargs
) -> pd.DataFrame:
    """
    Load a range from a Google Spreadsheet to a pandas DataFrame.

    Args:
        key: 44-digit spreadsheet id from the url e.g. 1DR1...k-xDU
        cell_range: data range including column names e.g. B1:C10
        sheet_name: name of the sheet within the spreadsheet e.g. Sheet1
        sheet_index: zero-based index where the sheet is within the spreadsheet
        **kwargs: any that are supported by this pandas version’s text parsing
            readers, such as pandas.read_csv

    Returns:
        df: pandas dataframe of the selected cell range

    Notes:
        - only one of sheet_name/sheet_index is required, defaults to the first
            sheet but prioritizing sheet_name if both are given

    Requirements:
        - share gs with the service account
        - gs api enabled for the project
    """
    # df filter params
    skiprows, nrows, usecols = lib._extract_cell_range_params(cell_range)

    df = load_gs_to_dataframe(
        key=key,
        sheet_name=sheet_name,
        sheet_index=sheet_index,
        skiprows=skiprows,
        nrows=nrows,
        usecols=usecols,
        **kwargs
    )
    return df


def load_gs_to_json_schema(
    key: str,
    schema_path: str,
    cell_range: str = None,
    sheet_name: str = "schema",
    sheet_index: int = 0,
):
    """
    Load table schema from google spreadsheet and write as json.

    Args:
        key: 44-digit spreadsheet id from the url e.g. '1DR1...k-xDU'
        schema_path: path to the schema file (the output to be written)
            e.g. 'path/schema.json'
        cell_range: data range including column names e.g. 'B1:C10'
        sheet_name: name of the sheet within the spreadsheet e.g. 'Sheet1'
        sheet_index: zero-based index where the sheet is within the spreadsheet
        column_order: the order of columns given for the schema, can only
            contain any of

    Notes:
        - if no 'cell_range' is defined, the sheet has to be otherwise empty
        - only one of sheet_name/sheet_index is required, defaults to the first
            sheet but prioritizing sheet_name if both are given
        - the gs load required the python data types otherwise it's making
            incorrect transformations
        - 'column mode' will be set as 'NULLABLE'

    Requirements:
        - share gs with the service account
        - gs api enabled for the project
    """
    # load gs
    if cell_range:
        df = load_gs_range_to_dataframe(
            key=key,
            cell_range=cell_range,
            sheet_name=sheet_name,
            sheet_index=sheet_index,
            evaluate_formulas=True
        )
    else:
        df = load_gs_to_dataframe(
            key=key,
            sheet_name=sheet_name,
            sheet_index=sheet_index,
            evaluate_formulas=True
        )
    # find header / drop empty rows and columns
    df = df.dropna(how="all", axis=1).dropna(how="all", axis=0)
    if any("Unnamed" in x for x in df.columns):
        df = df.rename(columns=df.iloc[0]).drop(df.index[0])

    # validate column names
    assert set(df.columns).issubset(["name", "type", "mode", "description"]), (
        "columns should be 'name', 'type', 'mode', 'description'"
    )
    assert set(["name", "type"]).issubset(list(df.columns)), (
        "'name' and 'type' are mandatory columns"
    )
    if "mode" not in df.columns:
        df["mode"] = "NULLABLE"
    else:
        df["mode"] = df["mode"].fillna("NULLABLE")
    if "description" not in df.columns:
        df["description"] = ""
    else:
        df["description"] = df["description"].fillna("")
    df = df[[
        "name",
        "type",
        "mode",
        "description"
    ]]
    # write
    f_path, f_name = split_full_path(schema_path)
    if f_path:
        create_folder(f_path)

    json_data = []
    for _, row in df.iterrows():
        json_data.append(
            {
                "name": row["name"],
                "type": row["type"],
                "mode": row["mode"],
                "description": row["description"]
            }
        )
    with open(schema_path, "w") as f:
        json.dump(json_data, f, indent=4)
        f.write("\n")
    print(f"Table schema file created at '{schema_path}.'")


def load_gs_to_sql_mock(
    key: str,
    mock_path: str,
    sheet_name: str = "acceptance",
    sheet_index: int = 1
):
    """
    Load mock data from google spreadsheet and write as cte sql.

    Args:
        key: 44-digit spreadsheet id from the url e.g. '1DR1...k-xDU'
        mock_path: path to the mock data file (the output for the cte)
            e.g. 'path/mock.sql'
        sheet_name: name of the sheet within the spreadsheet e.g. 'Sheet1'
        sheet_index: zero-based index where the sheet is within the spreadsheet

    Notes:
        - only one of sheet_name/sheet_index is required, defaults to the first
            sheet but prioritizing sheet_name if both are given
        - the gs load required the python data types otherwise it's making
            incorrect transformations

    Requirements:
        - share gs with the service account
        - gs api enabled for the project
    """
    # get table list and info
    table_columns = lib._extract_table_columns(
        key=key,
        sheet_name=sheet_name,
        sheet_index=sheet_index
    )
    # loop through tables
    count = 0
    ctes = []
    for table_name, cols in table_columns.items():

        # avoid quota errors - wait after every 5 tables
        if count != 0 and count % 5 == 0:
            sleep(15)
        count += 1

        # get schema
        schema_sql = lib._extract_table_schema(
            key=key,
            cols=cols,
            sheet_name=sheet_name,
            sheet_index=sheet_index
        )
        schema_py = lib._schema_sql_to_py(schema_sql)
        schema_sql = lib._schema_sql_to_bq_compatibility(schema_sql)

        # get table and convert to cte
        df = lib._extract_table_data(
            key=key,
            cols=cols,
            schema=schema_py,
            sheet_name=sheet_name,
            sheet_index=sheet_index
        )
        cte = lib._df_to_cte(
            df=df,
            schema_sql=schema_sql,
            cte_name=table_name
        )
        ctes.append(cte)

    # write
    f_path, f_name = split_full_path(mock_path)
    if f_path:
        create_folder(f_path)

    content = ",\n\n".join(ctes) + "\n"
    write_text_file(
        content=content,
        file_path=mock_path
    )
    print(f"Mock CTE file created at '{mock_path}.'")



""" Analytics Toolkit """
import errno
import hashlib
import hmac
import logging
import logging.config
import nltk
import os
import pandas as pd
import re
import yaml
from datetime import datetime, timedelta
from typing import List


class SafeDict(dict):
    def __missing__(self, key):
        return '{' + key + '}'


def read_sql_file(path_to_file, **kwargs):
    """ Loads SQL file as a string.
    Arguments:
        path_to_file (string): path to sql file from PROJECT_ROOT
        *kwargs: can be passed if some parameters are to be passed.

    Returns:
        a SQL formated with **kwargs if applicable.

    Example:
        With SQL as:
            "select .. {param2} .. {param1} .. {paramN}"
        *kwargs as:
            param1=value1
            param2=value2
            paranN=valueN
        The functions returns:
            "select .. value2 .. value1 .. valueN"
    """
    full_path = os.path.join(os.environ.get("PROJECT_ROOT"), path_to_file)

    file = open(full_path, 'r')
    sql = file.read()
    file.close()

    if len(kwargs) > 0:
        sql = sql.format_map(SafeDict(**kwargs))
    return sql


def read_yaml_file(path_to_file):
    """ Loads YAML file as a dictionary.
    Arguments:
    - path_to_file (string): path to yaml file from PROJECT_ROOT
    """
    full_path = os.path.join(os.getenv("PROJECT_ROOT"), path_to_file)
    with open(full_path, 'r') as file:
        config = yaml.safe_load(file)
    file.close()
    return config


def customer_hash(email):
    """ Hash Email address using sha256 algorithm with salt.
    Parameters:
    email (string): Email address of the customer
    Returns:
    UID (string)
   """
    if "EMAIL_HASH_SALT" in os.environ:
        pass
    else:
        raise KeyError("EMAIL_HASH_SALT does not exist")
    if isinstance(email, str):
        if email != '':
            email = bytes(email.lower(), 'utf-8')
            salt = bytes(os.environ.get("EMAIL_HASH_SALT"), 'utf-8')
            hash_ = hmac.new(key=salt,
                             digestmod=hashlib.sha256)
            hash_.update(email)
            uid = str(hash_.hexdigest())[0:16]
        else:
            uid = '0000000000000000'
        return uid
    elif email is None:
        uid = '0000000000000000'
        return uid
    else:
        raise KeyError("Email argument should be a string")


def stringify(value):
    """
    Returns the string representation of the value.
    """
    if value is None:
        return 'null'
    elif value is True:
        return 'True'
    elif value is False:
        return 'False'
    return str(value)


def is_email_address(text):
    """
    Return true if it is a valid email address
    """
    return re.search(r'[\w\.-]+@[\w\.-]+', text)


def anonymizer(text):
    """
    A part-of-speech tagger, or POS-tagger, processes a sequence
    of words, and attaches a part of speech tag to each word. See
    https://www.nltk.org/index.html for more information.
    Cant be used ATM. Need to add the process installation to base image.
    $ pip install -U textblob
    $ python -m textblob.download_corpora
    """
    if text is None or text == "":
        new_text = text
    else:
        new_text = []
        # Splits text into sentences
        sentence_list = text.replace('\n', ' ').split(". ")
        for sentence in sentence_list:
            # Splits sentence into list of words and filters empty elts.
            # Not using nltk.word_tokenize as it splits an email address
            # in several entities.
            word_list = list(filter(None, sentence.split(" ")))
            # process word_list
            pos = nltk.pos_tag(word_list)
            new_word_list = []
            for word in pos:
                if is_email_address(word[0]):
                    # tags word as EMAIL
                    new_word_list.append("{EMAIL}")
                elif word[1] == 'NNP':
                    # tags word as NAME (proper noun)
                    new_word_list.append("{NAME}")
                elif word[1] == 'CD':
                    # tags word as NUMBER
                    new_word_list.append("{NUMBER}")
                else:
                    # no tranformation
                    new_word_list.append(word[0])
            new_sentence = " ".join(new_word_list)
            new_text.append(new_sentence)
        new_text = ". ".join(new_text)
    return new_text


def get_date(window, date_format="%Y-%m-%d"):
    """
    Returns date in string format ('%Y-%m-%d') from today using window in days.
    get_date(window=0) returns today, get_date(widnow=1) returns yesterday...
    """
    date = datetime.today() - timedelta(days=window)
    return date.strftime(date_format)


def get_today_date(date_format="%Y-%m-%d"):
    """
    Returns today date in string format specified. Defaults to '%Y-%m-%d'.
    """
    return get_date(0, date_format)


def get_yesterday_date(date_format="%Y-%m-%d"):
    """
    Returns yesterday date in string format specified. Defaults to '%Y-%m-%d'.
    """
    return get_date(1, date_format)


def date_lister(start_date, end_date):
    """
     Returns list of dates between start_date and end_date in string format ('%Y-%m-%d')
     Arguments:
     - start_date (string)
     - end_date (string)
     """
    if end_date < start_date:
        date_list = []
        logging.error("End date must be equal or after start_date")
    else:
        date_list = pd.date_range(start_date, end_date)
        date_list = date_list.format()
        logging.info(date_list)
    return date_list


def validate_date(date, format='%Y-%m-%d', error_msg=None):
    try:
        datetime.strptime(date, format)
    except ValueError:
        if error_msg is None:
            raise ValueError("Incorrect data format, should be {}".format(format))
        else:
            raise ValueError(error_msg)


def level_is_valid(verbosity):
    '''
    Validates key parameter for configure_logging() and raises approriate
    exception messages.

    Returns the expected level for verbosity.
    '''

    if not isinstance(verbosity, str):
        raise ValueError('verbosity needs to be a string')
    verbosity = verbosity.upper()
    logging.info("verbosity passed: %s", verbosity)

    if verbosity not in ('DEBUG', 'INFO', 'QUIET', 'WARNING', 'CRITICAL'):
        raise ValueError('verbosity needs to be one of the following: DEBUG, INFO, QUIET, WARNING, CRITICAL')

    return verbosity


def configure_logging(env=None, verbosity=None):
    """ Sets logging """

    levels = {
        "DEBUG": {
            'level': "DEBUG",
            'format': "%(asctime)s,%(msecs)3d - %(levelname)-8s - %(funcName)5s:%(lineno)-10s :: %(message)s",
            'datefmt': "%Y-%m-%d %H:%M:%S"
        },
        "INFO": {
            'level': "INFO",
            'format': "%(asctime)s - %(levelname)-8s — %(name)-10s :: %(message)s",
            'datefmt': "%H:%M:%S"
        },
        "QUIET": {
            'level': "ERROR",
            'format': "%(asctime)s - %(levelname)-8s — %(name)-10s :: %(message)s",
            'datefmt': "%H:%M:%S"
        },
        "WARNING": {
            'level': "WARNING",
            'format': "%(asctime)s - %(levelname)s - %(funcName)-10s :: %(message)s",
            'datefmt': "%H:%M:%S"
        },
        "CRITICAL": {
            'level': "CRITICAL",
            'format': "%(asctime)s - %(levelname)-8s - %(funcName)5s:%(lineno)-10s :: %(message)s",
            'datefmt': "%Y-%m-%d %H:%M:%S"
        }
    }

    if verbosity is None:
        setlevel = levels["INFO"]['level']
        setformat = levels["INFO"]['format']
        setdatefmt = levels["INFO"]['datefmt']
    else:
        verbosity = level_is_valid(verbosity)
        setlevel = levels[verbosity]['level']
        setformat = levels[verbosity]['format']
        setdatefmt = levels[verbosity]['datefmt']

    logging.config.dictConfig(
        {
            'version': 1,
            'disable_existing_loggers': False,
            'root': {
                'level': setlevel,
                'handlers': ['default']
            },
            'formatters': {
                'default': {
                    'format': setformat,
                    'datefmt': setdatefmt
                },
            },
            'handlers': {
                'default': {
                    'class': 'logging.StreamHandler',
                    'stream': 'ext://sys.stdout',
                    'formatter': 'default',
                    'level': logging.DEBUG
                }
            }
        }
    )


def convert_list_for_sql(my_list):
    """ Convert a python list to a SQL list.
    The function is primarly used when trying to format SQL queries by passing an argument.

    Arguments:
        my_list: list of elements to be used in a SQL query

    Example:
        1. convert_list_for_sql([1, 2, 3]) returns '1, 2, 3'
        2. convert_list_for_sql(['Simone', 'Dan']) returns ''Simone', 'Dan''
    """
    final_list = []
    for item in my_list:
        if isinstance(item, str):
            _item = '\'{}\''.format(item)
        else:
            _item = item
        final_list.append(_item)
    return ", ".join([str(item) for item in final_list])


def split_full_path(
    path: str
) -> List[str]:
    """
    Splits a full file path to path and file name.

    Args:
        path: full file path e.g. 'path/to/file.json'

    Returns:
        [f_path, f_name]: ['path/to', 'file.json']
    """
    split = path.split("/")
    f_name = split[-1]
    split.remove(f_name)
    f_path = "/".join(split)

    return f_path, f_name


def create_folder(
    path,
    print_status: bool = True
):
    """
    Checks if a folder exists, if not creates it.

    Args:
        path: folder path
        print_status: print folder status
    """
    try:
        os.makedirs(path)
        if print_status:
            print(f"Folder created at '{path}'.")
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def write_text_file(
    content: str,
    file_path: str
):
    """
    Write string to a text file.

    Args:
        content: string to write
        file_path: e.g. 'path/file.txt'
    """
    t = open(file_path, "w")
    t.write(content)
    t.close()


def slice_list_to_chunks(lst, n):
    """
    Slice a list into chunks of size n.

    Args:
        list
        int
    Returns:
        [list]
    """
    chunks = [lst[x:x+n] for x in range(0, len(lst), n)]

    return chunks


def flatten_df_columns(df: pd.DataFrame) -> pd.DataFrame:
    """
    Flatten repeated fields in a pandas dataframe an unnest columns.
    (for single layer nesting only)
    Checks if column is nested - if not raises AttributeError and appends
    original column.

    Args:
        df
    Returns:
        df
    """
    # explode repeated fields
    for col in df.columns:
        df = df.explode(col)
    df = df.reset_index(drop=True)

    # unnest
    df_flat = pd.DataFrame()
    for col in df.columns:
        try:
            col_flat = df[col].apply(lambda x: {} if pd.isna(x) else x)
            col_flat = pd.json_normalize(col_flat)
            if col_flat.shape[1] == 1:
                raise AttributeError
            else:
                col_flat.columns = [f"{col}__{x}" for x in col_flat.columns]
                df_flat = pd.concat([df_flat, col_flat], axis=1)
        except AttributeError:
            df_flat = pd.concat([df_flat, df[col]], axis=1)
    df_flat = df_flat.reset_index(drop=True)

    return df_flat


def get_env_var(
        env_var_name: str,
        break_if_missing: bool = False,
        allowed_values: list = []
) -> str:
    """
    Returns env var, optional check to see if exists and optional check against
    allowed values.
    """
    env_var = os.getenv(env_var_name)

    if break_if_missing and not env_var:
        raise ValueError(f"Env var '{env_var_name}' doesn't exist")

    if allowed_values:
        assert env_var in allowed_values, (
            f"Env var value for '{env_var_name}' is not in allowed values "
            f"list '{allowed_values}'"
        )

    return env_var
